//
//  Copyright (c) 2003 Daniel Frey
//
//  Use, modification and distribution are subject to the Boost Software License, Version 1.0
//
//  (See accompanying file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
//

#include <iostream>
#include <cmath>  // for sqrt


// #include "my_constants.hpp" from includes constant.hpp

namespace boost
{
    template< template< typename > class F > struct constant
    {
        template< typename T > typename F< T >::result_type get() const
        {
            return F< T >::get();
        }
    };
}

namespace boost
{
    namespace config
    {
        namespace constant
        {
            // We want to allow direct comparison of constants against integers.
            // We use the conversion to double to get the constant's value.
            // Later, we will of course provide more comparisons to other integral
            // types like unsigned, long, etc. as required.
            template<> struct map_compared_type< int >  // <<<<<<<<<<<<<<<<<<< c2143 here
						{
							typedef double type;
						};
/*
						syntax error : missing 'token1' before 'token2'

The compiler expected a specific token (a language element other than white space) and found another token instead. 

Possible causes 


Missing closing brace, parenthesis, or semicolon on the line where the error is detected or a few lines above: 
// C2143c.cpp
class X
{
   int member
} x;          // C2143, missing ; on previous line

Invalid tag in a class declaration: 
// C2143d.cpp
class X
{
   int member;
} x;
class + {};   // C2143, + is an invalid tag name

A label not attached to a statement. If you must place a label by itself (for example, at the end of a compound statement), attach it to a null statement: 
// C2143e.cpp
void func1()
{
   end:      // C2143
}            // this line is not a statement
void func2()
{
   end:      // OK
    ;        // this label attached to null statement
}

Invalid attribute syntax. Although the following instruction is valid in a MIDL file: 
typedef [public] long MEMBERID;
it is not valid in a C++ source code file. Instead, you would need to issue: 

[public] typedef long MEMBERID;
See Knowledge Base article Q241706 for information on this error when using a function-try block.
/*

            // We use this as a fallback.
            // If conversion to std::complex is requested from the user, 
            // but not specified for a constant, the constant is asked to
            // provide T. This is then used to construct the std::complex< T >.
            // Neither the constant framework nor the constants you include need
            // to know anything (or include) std::complex. That way, the type
            // is nicely added only when the users wants/needs it.
      //      template< typename T > struct map_type< std::complex< T > >
						//{
						//	typedef T type;
						//};
        } // namespace constant
    } // namespace config
}   // namespace boost


namespace boost
{
    namespace math
    {
        // A first, simple example: pi
        template< typename T > struct pi_value : constant_value< math::pi_value, T > {};
        struct pi_t : constant< pi_value > {} const pi = pi_t();

        template<> struct pi_value< float > {
            typedef float result_type;
            static result_type get() { return 3.14F; }
        };
        template<> struct pi_value< double > { 
            typedef double result_type;
            static result_type get() { return 3.1416; }
        };
        template<> struct pi_value< long double > { 
            typedef long double result_type;
            static result_type get() { return 3.1415927L; }
        };

        //// A solution to the naming dilemma, providing pi^2.
        //// Constructed (or combined) constants follow a slightly different scheme.
        //// Note that the calculation is done at run-time in case the constant is not explicitly specialized for a type.
        //template< typename T > struct pi2_value { 
        //    typedef typename pi_value< T >::result_type result_type;
        //    static result_type get() { return pi.get< T >() * pi.get< T >(); }
        //};
        //// We do not provide a constant 'pi2', instead we
        //// provide 'pi*pi' to access pi2_value.
        //inline constant< pi2_value > operator*( const constant< pi_value >&, const constant< pi_value >& ) { return constant< pi2_value >(); }

        //template<> struct pi2_value< float > { 
        //    typedef float result_type;
        //    static result_type get() { return 9.87; }
        //};
        //template<> struct pi2_value< double > { 
        //    typedef double result_type;
        //    static result_type get() { return 9.8696; }
        //};
        //template<> struct pi2_value< long double > { 
        //    typedef long double result_type;
        //    static result_type get() { return 9.8696044; }
        //};
    }

    // Some obvious constants
#define CONSTANT_VALUE( name, value ) \
    template< typename T > struct name##_value { typedef T result_type; static result_type get() { return value; } }; \
    struct name##_t : constant< name##_value > {} const name = name##_t()

    CONSTANT_VALUE( minus_one, -1 );
    CONSTANT_VALUE( zero, 0 );
    CONSTANT_VALUE( one, 1 );
    CONSTANT_VALUE( two, 2 );
    CONSTANT_VALUE( three, 3 );
    CONSTANT_VALUE( ten, 10 );
#undef CONSTANT_VALUE

    // Another example to solve the naming dilemma, this time providing 'sqrt( two )'
    template< typename T > struct sqrt2_value { 
        typedef typename two_value< T >::result_type result_type;
        static result_type get() { return std::sqrt( two.get< T >() ); }
    };
    template<> struct sqrt2_value< float > { 
        typedef float result_type;
        static result_type get() { return 1.4142; }
    };
    template<> struct sqrt2_value< double > { 
        typedef double result_type;
        static result_type get() { return 1.41421356; }
    };
    template<> struct sqrt2_value< long double > {
        typedef long double result_type;
        static result_type get() { return 1.4142135623730950488016887242096980785697L; }
    };

    inline constant< sqrt2_value > sqrt( const constant< two_value >& ) { return constant< sqrt2_value >(); }

    //// Or you create a complex-only constant
    //template< typename > struct i_value;
    //template< typename T > struct i_value< std::complex< T > > { 
    //    typedef std::complex< T > result_type;
    //    static result_type get() { return result_type( 0, 1 ); }
    //};
    //struct i_t : constant< i_value > {} const i = i_t();
}

int main()
{
    // Usage example:

    const float f = 1.F;
    const double d = 1.;
    const long double ld = 1.L;

    using boost::math::pi;

    // explicit casting
    std::cout << pi.get< float >() << std::endl;
    std::cout << pi.get< double >() << std::endl;
    std::cout << pi.get< long double >() << std::endl;

    // implicit casting
    const float f2 = pi;
    const double d2 = pi;
    const long double ld2 = pi;
    
    std::cout << f2 << std::endl;
    std::cout << d2 << std::endl;
    std::cout << ld2 << std::endl;
     /*

    std::cout << pi + 2. << std::endl;
    std::cout << 2. + pi << std::endl;

    std::cout << pi * f * f << std::endl;
    std::cout << pi * d * d << std::endl;
    std::cout << pi * ld * ld << std::endl;

    // Using pi*pi, which is not calculated at runtime
    // but defined as a single, fast, high-precision
    // constant on its own. (OK, it *is* calculated at
    // run-time if not specialized for the type you are
    // using)
    std::cout << pi * pi * f << std::endl;
    std::cout << pi * pi * d << std::endl;
    std::cout << pi * pi * ld << std::endl;

    // Defered casting: pi^3 is not a constant.
    // Instead, pi^2 (which is a constant) is multiplied
    // with pi at runtime, and then again with f/d/ld.
    std::cout << pi * pi * pi * f << std::endl;
    std::cout << pi * pi * pi * d << std::endl;
    std::cout << pi * pi * pi * ld << std::endl;

    std::cout << pi + pi - pi * pi / pi + d << std::endl;
    
    // A more complex example:

    using std::sqrt;
    using boost::one;
    using boost::two;
    
    std::cout << sqrt( two ) * sqrt( 2.F ) << std::endl;
    std::cout << sqrt( two ) * sqrt( 2. ) << std::endl;
    std::cout << sqrt( two ) * sqrt( 2.L ) << std::endl;

    std::cout << sqrt( sqrt( two + pi ) ) + d << std::endl;

    //// std::complex
    //
    //using boost::i;

    //const std::complex< double > cd1( 1, 0 );
    //const std::complex< double > cd2 = cd1 * i;
    //
    //std::cout << cd1 << std::endl;
    //std::cout << cd2 << std::endl;

    //std::cout << cd1 * pi << std::endl;

    //std::cout << pi + i + pi * ( pi + i ) + i * ( pi * pi * cd1 ) << std::endl;

    //std::cout << pi.get< std::complex< double > >() << std::endl;

    //if( !( 1 == pi ) && 
    //    1 < pi &&
    //    !( 1 == pi ) &&
    //    1 != pi &&
    //    pi > 1 &&
    //    !( pi <= 1 ) &&
    //    !( 1 >= pi ) ) std::cout << "X" << std::endl;
		   */

} // int main()


/*

Most cut down version - as currently commented out.

------ Build started: Project: Frey constants3, Configuration: Debug Win32 ------

Compiling...
test_constants3.cpp
test_constants3.cpp(36) : error C2143: syntax error : missing ';' before '<'
test_constants3.cpp(36) : error C2913: explicit specialization; 'boost::config::constant::map_compared_type' is not a specialization of a class template
test_constants3.cpp(36) : error C2059: syntax error : '<'
test_constants3.cpp(37) : error C2143: syntax error : missing ';' before '{'
test_constants3.cpp(37) : error C2447: '{' : missing function header (old-style formal list?)
test_constants3.cpp(62) : error C2504: 'constant_value' : base class undefined
        test_constants3.cpp(62) : see reference to class template instantiation 'boost::math::pi_value<T>' being compiled
test_constants3.cpp(62) : error C2143: syntax error : missing ',' before '<'
test_constants3.cpp(161) : error C2440: 'initializing' : cannot convert from 'const boost::math::pi_t' to 'const float'
        No user-defined-conversion operator available that can perform this conversion, or the operator cannot be called
test_constants3.cpp(162) : error C2440: 'initializing' : cannot convert from 'const boost::math::pi_t' to 'const double'
        No user-defined-conversion operator available that can perform this conversion, or the operator cannot be called
test_constants3.cpp(163) : error C2440: 'initializing' : cannot convert from 'const boost::math::pi_t' to 'const long double'
        No user-defined-conversion operator available that can perform this conversion, or the operator cannot be called

Build Time 0:00
Build log was saved at "file://j:\Cpp\Frey constants3\Debug\BuildLog.htm"
Frey constants3 - 10 error(s), 0 warning(s)


---------------------- Done ----------------------

    Build: 0 succeeded, 1 failed, 0 skipped


------ Build started: Project: Frey constants3, Configuration: Debug Win32 ------

Compiling...
test_constants3.cpp
test_constants3.cpp(36) : error C2143: syntax error : missing ';' before '<'
test_constants3.cpp(36) : error C2913: explicit specialization; 'boost::config::constant::map_compared_type' is not a specialization of a class template
test_constants3.cpp(36) : error C2059: syntax error : '<'
test_constants3.cpp(37) : error C2143: syntax error : missing ';' before '{'
test_constants3.cpp(37) : error C2447: '{' : missing function header (old-style formal list?)
test_constants3.cpp(62) : error C2504: 'constant_value' : base class undefined
        test_constants3.cpp(62) : see reference to class template instantiation 'boost::math::pi_value<T>' being compiled
test_constants3.cpp(62) : error C2143: syntax error : missing ',' before '<'
test_constants3.cpp(171) : error C2676: binary '+' : 'const boost::math::pi_t' does not define this operator or a conversion to a type acceptable to the predefined operator
test_constants3.cpp(172) : error C2677: binary '+' : no global operator found which takes type 'const boost::math::pi_t' (or there is no acceptable conversion)
test_constants3.cpp(174) : error C2679: binary '*' : no operator found which takes a right-hand operand of type 'const float' (or there is no acceptable conversion)
test_constants3.cpp(175) : error C2679: binary '*' : no operator found which takes a right-hand operand of type 'const double' (or there is no acceptable conversion)
test_constants3.cpp(176) : error C2679: binary '*' : no operator found which takes a right-hand operand of type 'const long double' (or there is no acceptable conversion)
test_constants3.cpp(183) : error C2678: binary '*' : no operator found which takes a left-hand operand of type 'boost::constant<F>' (or there is no acceptable conversion)
        with
        [
            F=boost::math::pi2_value
        ]
test_constants3.cpp(184) : error C2678: binary '*' : no operator found which takes a left-hand operand of type 'boost::constant<F>' (or there is no acceptable conversion)
        with
        [
            F=boost::math::pi2_value
        ]
test_constants3.cpp(185) : error C2678: binary '*' : no operator found which takes a left-hand operand of type 'boost::constant<F>' (or there is no acceptable conversion)
        with
        [
            F=boost::math::pi2_value
        ]
test_constants3.cpp(190) : error C2678: binary '*' : no operator found which takes a left-hand operand of type 'boost::constant<F>' (or there is no acceptable conversion)
        with
        [
            F=boost::math::pi2_value
        ]
test_constants3.cpp(191) : error C2678: binary '*' : no operator found which takes a left-hand operand of type 'boost::constant<F>' (or there is no acceptable conversion)
        with
        [
            F=boost::math::pi2_value
        ]
test_constants3.cpp(192) : error C2678: binary '*' : no operator found which takes a left-hand operand of type 'boost::constant<F>' (or there is no acceptable conversion)
        with
        [
            F=boost::math::pi2_value
        ]
test_constants3.cpp(194) : error C2676: binary '+' : 'const boost::math::pi_t' does not define this operator or a conversion to a type acceptable to the predefined operator
test_constants3.cpp(194) : error C2676: binary '/' : 'boost::constant<F>' does not define this operator or a conversion to a type acceptable to the predefined operator
        with
        [
            F=boost::math::pi2_value
        ]
test_constants3.cpp(202) : error C2676: binary '*' : 'boost::constant<F>' does not define this operator or a conversion to a type acceptable to the predefined operator
        with
        [
            F=boost::sqrt2_value
        ]
test_constants3.cpp(203) : error C2676: binary '*' : 'boost::constant<F>' does not define this operator or a conversion to a type acceptable to the predefined operator
        with
        [
            F=boost::sqrt2_value
        ]
test_constants3.cpp(204) : error C2676: binary '*' : 'boost::constant<F>' does not define this operator or a conversion to a type acceptable to the predefined operator
        with
        [
            F=boost::sqrt2_value
        ]
test_constants3.cpp(206) : error C2676: binary '+' : 'const boost::two_t' does not define this operator or a conversion to a type acceptable to the predefined operator

Build Time 0:00
Build log was saved at "file://j:\Cpp\Frey constants3\Debug\BuildLog.htm"
Frey constants3 - 24 error(s), 0 warning(s)


---------------------- Done ----------------------

    Build: 0 succeeded, 1 failed, 0 skipped


*/